@@ -2,8 +2,8 @@
 
 It does not require a server running.
 """
-from .container import Container
+from .container import Container, ObjectType
 
-__all__ = ('Container',)
+__all__ = ('Container', 'ObjectType')
 
-__version__ = '0.3.0'
+__version__ = '0.4.0'
