@@ -4,13 +4,25 @@ Some might be useful also for end users, like the wrappers to get streams,
 like the ``LazyOpener``.
 """
 import hashlib
+import itertools
 import os
 import uuid
 import zlib
 
+try:
+    import fcntl
+except ImportError:
+    # Not available on Windows
+    fcntl = None
+
 from contextlib import contextmanager
 
-from .exceptions import ModificationNotAllowed, ClosingNotAllowed, DynamicInconsistentContent
+from .exceptions import ModificationNotAllowed, ClosingNotAllowed
+
+# For now I I don't always activate it as I need to think at the right balance between
+# safety and performance/disk wearing
+# I use it only when storing packs
+_MACOS_ALWAYS_USE_FULLSYNC = False
 
 
 class LazyOpener:
@@ -79,7 +91,8 @@ def nullcontext(enter_result):
 class ObjectWriter:
     """A class to get direct write access for a new object."""
 
-    def __init__(self, sandbox_folder, loose_folder, loose_prefix_len, hash_type, trust_existing=False):
+    def __init__(  # pylint: disable=too-many-arguments
+            self, sandbox_folder, loose_folder, loose_prefix_len, duplicates_folder, hash_type, trust_existing=False):
         """Initialise an object to store a new loose object.
 
         :param sandbox_folder: the folder to store objects while still giving
@@ -90,11 +103,14 @@ class ObjectWriter:
           0 means store objects flat. Note that objects are always stored flat
           in the sandbox (there shouldn't be too many sandbox objects at the
           same time).
+        :param duplicates_folder: the folder to store the duplicates objects (for concurrent writes
+          on Windows where I cannot replace).
         :param trust_existing: if True, just continues if a file with the same hashkey is found.
           Otherwise, check the existing file content and overwrite if wrong.
         """
         self._sandbox_folder = sandbox_folder
         self._loose_folder = loose_folder
+        self._duplicates_folder = duplicates_folder
         self._hash_type = hash_type
         self._hashkey = None
         self._loose_prefix_len = loose_prefix_len
@@ -122,14 +138,14 @@ class ObjectWriter:
         if self._filehandle is not None:
             raise IOError('You have already opened this ObjectWriter instance')
         if self._stored:
-            raise ModificationNotAllowed("You have already stored this object '{}'".format(self.get_hashkey()))
+            raise ModificationNotAllowed("You have already tried to store this object '{}'".format(self.get_hashkey()))
         # Create a new uniquely-named file in the sandbox.
         # It seems faster than using a NamedTemporaryFile, see benchmarks.
         self._obj_path = os.path.join(self._sandbox_folder, uuid.uuid4().hex)
         self._filehandle = HashWriterWrapper(open(self._obj_path, 'wb'), hash_type=self.hash_type)
         return self._filehandle
 
-    def __exit__(self, exc_type, value, traceback):  # pylint: disable=too-many-branches
+    def __exit__(self, exc_type, value, traceback):  # pylint: disable=too-many-branches, too-many-statements
         """
         Close the file object, and move it from the sandbox to the loose
         object folder, possibly using sharding if loose_prexix_len is not 0.
@@ -138,10 +154,10 @@ class ObjectWriter:
             if exc_type is None:
                 if self._filehandle.closed:
                     raise ClosingNotAllowed('You cannot close the file handle yourself!')
-                # Flush out of the buffer, then fsync so it's written to disk; see e.g.
-                # https://blog.gocept.com/2013/07/15/reliable-file-updates-with-python/
-                self._filehandle.flush()
-                os.fsync(self._filehandle.fileno())
+                # Flush out of the buffer and sync to disk so data is preserved even for power failures
+                # NOTE: For now I don't use `fullsync` on Mac for performance reasons, for loose objects - to decide
+                # if we want to change this!
+                safe_flush_to_disk(self._filehandle, self._obj_path)
                 self._filehandle.close()
                 self._hashkey = str(self._filehandle.hexdigest())
                 self._filehandle = None
@@ -164,70 +180,133 @@ class ObjectWriter:
                     dest_loose_object = os.path.join(self._loose_folder, self._hashkey)
 
                 dest_parent_folder = os.path.dirname(dest_loose_object)
+                exists_wrong_checksum = False
                 # At this point, if the destination exists, it means someone already put an object
                 # with the same content/hash key
                 if os.path.exists(dest_loose_object):
                     if self._trust_existing:
                         # I trust that the object is correct: I just return
+                        # Note: if another process is deleting the file at the same time (right after),
+                        # it will be deleted. But this is OK - the deletion is in an independent process,
+                        # an if it happens only microseconds or seconds after the write, the effect is the same:
+                        # the object is deleted and the caller of this function will not find it anymore.
                         return
                     # I do not trust the object is correct.
                     # This might still not be perfect: while I check, another
                     # process might in the meantime rewrite it again, and this might
-                    # be wrong. But this is very difficult to catch
-                    existing_checksum = _compute_hash_for_filename(filename=dest_loose_object, hash_type=self.hash_type)
+                    # be wrong. But this is very difficult to catch, and in general
+                    # the situation in which a process writes a corrupt node is really an error
+                    try:
+                        existing_checksum = _compute_hash_for_filename(
+                            filename=dest_loose_object, hash_type=self.hash_type
+                        )
+                    except PermissionError:
+                        # On Windows I might get a PermissionError. I store a copy and return.
+                        # This would happen if e.g. the file exists but is being moved in place and
+                        # I cannot open for reading.
+                        self._store_duplicate_copy(self._obj_path, self._hashkey)
+                        return
                     if existing_checksum == self._hashkey:
                         # The existing object has the correct hash, I just return.
                         return
-                    # The existing object has the wrong hash! I am going to assume where was a problem
-                    # and I will just remove it and overwrite it.
-                    # This means that the object will *not* be there for some time.
-                    # But this should be ok, because it was corrupted, so nobody should really be
-                    # using it...
-                    # Note: this would really be needed only on Windows, on Unix os.rename will silently replace
-                    # the existing file
-                    # Note that in reality I could end up deleting a 'good' file written by another process
-                    # performing the exact same operation. But I will write it again below so this will only
-                    # create a small window of time in which it does not exist (and I don't know if there is
-                    # a way around it). This anyway happens only for existing but corrupted loose objects.
-                    os.remove(dest_loose_object)
-                # This is an atomic operation, at least according to this website:
-                # https://alexwlchan.net/2019/03/atomic-cross-filesystem-moves-in-python/
-                # but needs to be on the same filesystem (this should always be the case for us)
-                # Note that instead shutil.move is not guaranteed to be atomic!
-                # Also: on Linux, this performs a silent overwrite if the file exists.
-                # Instead, it raises OSError on Windows if the file exists.
-                # The Linux behavior is ok: in this case, two processes were writing at almost the same time;
-                # we assume it's ok to take either of them (the second winning)
-                # On Windows, we need to take special action
-                try:
-                    os.rename(self._obj_path, dest_loose_object)
-                except OSError:
-                    # NOTE! This branch only happens on Windows, see above
-                    # A file with the same name was written in the meantime by someone else...
-                    # We do a final check of its content
-                    existing_checksum = _compute_hash_for_filename(filename=dest_loose_object, hash_type=self.hash_type)
-                    if existing_checksum == self._hashkey:
-                        # The existing object that has appeared in the meantime is OK. Fine, has the correct hash,
-                        # I just return.
+                    # If existing_checksum is None, the file has been removed in the meantime.
+                    # It could be a delete of the object: then, for consistency with the logic above,
+                    # I decide that the two operations that happened almost at the same time could
+                    # have happened in swapped order, and I don't write it back.
+                    # Or it's a packing operation (and then it's OK to not put the file back).
+                    # I therefore just return
+                    if existing_checksum is None:
                         return
-                    raise DynamicInconsistentContent(
-                        "I am trying to create loose object with hash key '{}', "
-                        'but it keeps reappearing with wrong content!'.format(self._hashkey)
-                    )
+                    # If we are here, the file exists and has the wrong checksum. I mark this condition
+                    exists_wrong_checksum = True
+
+                # If I am here, there are two options:
+                # 1. The object did not exist
+                # 2. The file is there but its content is wrong. In this case I want to overwrite the object.
+
+                # I therefore call a 'replace' call, that should be an atomic operation
+                # Notes (on os.replace vs os.rename):
+                # - os.rename() on Linux is atomic (see e.g. also
+                #   https://alexwlchan.net/2019/03/atomic-cross-filesystem-moves-in-python/
+                #   but needs to be on the same filesystem (this should always be the case for us)
+                # - Remember that instead shutil.move is not guaranteed to be atomic!
+                # - os.rename performs a silent overwrite if the file exists on Posix, so would be enough
+                #   on Linux/Mac; instead, it raises OSError on Windows if the file exists.
+                # - we use os.rename: we make sure in this way, on Windows, that only one process writes
+                #   to the destination, avoiding race conditions with two processes trying to write the
+                #   same file.
+                if exists_wrong_checksum:
+                    # In this case I try a replace of the file
+                    try:
+                        os.replace(self._obj_path, dest_loose_object)
+                    except PermissionError:
+                        # NOTE! This branch only happens on Windows, when the
+                        # file with the same name is open by someone else...
+                        # I just store a duplicate copy. This happens if I find a file, it has a wrong
+                        # hash key, but then I cannot replace it (e.g. it's open, or two processes are writing
+                        # at the same time).
+                        self._store_duplicate_copy(self._obj_path, self._hashkey)
+                else:
+                    # In this case the file does not exist. I try just a rename of the file,
+                    # to put it in place and reduce the risk of multiple processes trying to overwrite
+                    # the same file
+                    try:
+                        os.rename(self._obj_path, dest_loose_object)
+                    except FileExistsError:
+                        # NOTE! This branch only happens on Windows, when the
+                        # file with the same name was opened in the meantime by someone else...
+                        if self._trust_existing:
+                            # If I trust existing files, I just return - someone put the file in place
+                            return
+                        # I'm here: the file did not exist, but appeared exactly at the same time: someone
+                        # else is writing it.
+                        # I do a final attempt to check the checksum! However, there is a high chance
+                        # that the file is being still locked for reading, so I need to take care of checking for those
+                        # error
+                        try:
+                            existing_checksum = _compute_hash_for_filename(
+                                filename=dest_loose_object, hash_type=self.hash_type
+                            )
+                            if existing_checksum == self._hashkey:
+                                # The existing object has the correct hash, I just return.
+                                return
+                        except PermissionError:
+                            # The file is probably being moved and I cannot open for reading.
+                            # I just pass, so in the lines below I will store a duplicate copy.
+                            pass
+
+                        # I'm here. Either I got a permission error, or the checksum is different (
+                        # and the file was created exactly in a concurrent process, it wasn't existing at the
+                        # beginning of this function): I just store a duplicate
+                        self._store_duplicate_copy(self._obj_path, self._hashkey)
 
                 # Flush also the parent directory, see e.g.
                 # https://blog.gocept.com/2013/07/15/reliable-file-updates-with-python/
+                # I do not call safe_flush_to_disk as I need to flush only the folder metadata (I think)
+                # Otherwise I should open again the file
                 if os.name == 'posix':
                     dirfd = os.open(os.path.dirname(dest_parent_folder), os.O_DIRECTORY)
                     os.fsync(dirfd)
                     os.close(dirfd)
-                self._stored = True
         finally:
+            # I set the stored flag, even if there was a problem, to avoid reuse of the same object.
+            self._stored = True
+
             if self._filehandle is not None and not self._filehandle.closed:
                 self._filehandle.close()
             if os.path.exists(self._obj_path):
                 os.remove(self._obj_path)
 
+    def _store_duplicate_copy(self, source_file, hashkey):
+        """This function is called (on Windows) when trying to store a file that already exists.
+
+        In the `clean_storage` I will clean up old copies if the hash matches.
+        """
+        # Destination file starts with the hashkey, then has an dot as a separator, and is
+        # followed by un UUID to make sure there is never a collision
+        dest_file = os.path.join(self._duplicates_folder, '{}.{}'.format(hashkey, uuid.uuid4().hex))
+        os.rename(source_file, dest_file)
+
 
 class PackedObjectReader:
     """A class to read from a pack file.
@@ -235,18 +314,43 @@ class PackedObjectReader:
     This ensures that the .read() method works and does not go beyond the
     length of the given object."""
 
+    @property
+    def mode(self):
+        return self._fhandle.mode
+
     @property
     def seekable(self):
         """Return whether object supports random access."""
-        return False
+        return True
 
-    def seek(self, target, whence=0):  # pylint: disable=no-self-use
-        """Change stream position."""
-        raise OSError('Object not seekable')
+    def seek(self, target, whence=0):
+        """Change stream position.
 
-    def tell(self):  # pylint: disable=no-self-use
-        """Return current stream position."""
-        raise OSError('Object not seekable')
+        Note that contrary to a standard file, also seeking beyond the borders will raise a ValueError.
+
+        :raises NotImplementedError: if ``whence`` is not 0 or 1.
+        """
+        if whence not in [0, 1]:
+            raise NotImplementedError('Invalid value for `whence`: only 0 and 1 are currently implemented.')
+
+        if whence == 1:
+            target = self.tell() + target
+
+        if target < 0:
+            raise ValueError('specified target would exceed the lower boundary of bytes that are accessible.')
+        if target > self._length:
+            raise ValueError('specified target would exceed the upper boundary of bytes that are accessible.')
+        new_pos = self._offset + target
+        self._fhandle.seek(new_pos)
+        # Next function MUST be called every time we move into the _fhandle file, to update the position
+        self._update_pos()
+
+        # seek returns the new absolute position
+        return target
+
+    def tell(self):
+        """Return current stream position, relative to the internal offset."""
+        return self._fhandle.tell() - self._offset
 
     def __init__(self, fhandle, offset, length):
         """
@@ -325,6 +429,7 @@ class StreamDecompresser:
         self._compressed_stream = compressed_stream
         self._decompressor = zlib.decompressobj()
         self._internal_buffer = b''
+        self._pos = 0
 
     def read(self, size=-1):
         """
@@ -381,9 +486,59 @@ class StreamDecompresser:
         # Note that we could be here also with len(self._internal_buffer) < size,
         # if we used 'break' because the internal buffer reached EOF.
         to_return, self._internal_buffer = self._internal_buffer[:size], self._internal_buffer[size:]
+        self._pos += len(to_return)
 
         return to_return
 
+    @property
+    def seekable(self):
+        """Return whether object supports random access."""
+        return True
+
+    def tell(self):
+        """Return current position in file."""
+        return self._pos
+
+    def seek(self, target, whence=0):
+        """Change stream position.
+
+        ..note:: This is particularly inefficient if `target > 0` since it will have
+           to decompress again from the beginning. So use with care!
+
+        :raises NotImplementedError: if ``whence`` is not 0 or 1.
+        """
+        read_chunk_size = 256 * 1024
+
+        if whence not in [0, 1]:
+            raise NotImplementedError('Invalid value for `whence`: only 0 and 1 are currently implemented.')
+
+        if whence == 1:
+            target = self.tell() + target
+
+        if target < 0:
+            raise ValueError('negative seek position {}'.format(target))
+        if target == 0:
+            # Going back to zero it's efficient. I need to reset all internal variables, as in the init.
+            self._compressed_stream.seek(0)
+            self._decompressor = zlib.decompressobj()
+            self._internal_buffer = b''
+            self._pos = 0
+            return 0
+
+        if target < self.tell():
+            # I cannot go backwards. In this case, I am forced to go back to zero and restart
+            # (I always know how to go back to zero). Otherwise, I just continue from where I am.
+            self.seek(0)
+
+        # Read target bytes, but at most `read_chunk_size` at a time to avoid memory overflow
+        while self.tell() < target:
+            content = self.read(min(read_chunk_size, target - self.tell()))
+            if not content:
+                # If I am asking a position beyond the end, I stop to avoid infinite loops
+                break
+        # Differently than files, I return here the actual position
+        return self._pos
+
 
 class ZeroStream:
     """A class to return an (unseekable) stream returning only zeros, with length length."""
@@ -427,13 +582,13 @@ class ZeroStream:
 def is_known_hash(hash_type):
     """Return True if the hash_type is known, False otherwise."""
     try:
-        _get_hash(hash_type)
+        get_hash(hash_type)
         return True
     except ValueError:
         return False
 
 
-def _get_hash(hash_type):
+def get_hash(hash_type):
     """Return a hash class with an update method and a hexdigest method."""
     known_hashes = {
         'sha256': hashlib.sha256,
@@ -451,20 +606,23 @@ def _compute_hash_for_filename(filename, hash_type):
     Will read the file in chunks.
 
     :param filename: a filename to a file to check.
-    :param hash_type: a valid string as recognised by the _get_hash function
-    :return: the hash hexdigest (the hash key)
+    :param hash_type: a valid string as recognised by the get_hash function
+    :return: the hash hexdigest (the hash key), or `None` if the file does not exist
     """
     _chunksize = 524288
 
-    hasher = _get_hash(hash_type)()
-    with open(filename, 'rb') as fhandle:
-        while True:
-            chunk = fhandle.read(_chunksize)
-            hasher.update(chunk)
+    hasher = get_hash(hash_type)()
+    try:
+        with open(filename, 'rb') as fhandle:
+            while True:
+                chunk = fhandle.read(_chunksize)
+                hasher.update(chunk)
 
-            if not chunk:
-                # Empty returned value: EOF
-                break
+                if not chunk:
+                    # Empty returned value: EOF
+                    break
+    except FileNotFoundError:
+        return None
 
     return hasher.hexdigest()
 
@@ -485,7 +643,7 @@ class HashWriterWrapper:
         assert 'b' in self._write_stream.mode
         self._hash_type = hash_type
 
-        self._hash = _get_hash(self._hash_type)()
+        self._hash = get_hash(self._hash_type)()
         self._position = self._write_stream.tell()
 
     @property
@@ -543,3 +701,88 @@ class HashWriterWrapper:
     def fileno(self):
         """Return the integer file descriptor of the underlying file object."""
         return self._write_stream.fileno()
+
+
+def chunk_iterator(iterator, size):
+    """Given an iterator, split it in chunks of size `size` (except the last one, that can be shorter).
+
+    Examples:
+
+    - `list(chunk_iterator(range(10), 3))` returns `[(0, 1, 2), (3, 4, 5), (6, 7, 8), (9,)]`
+    - `list(chunk_iterator(range(9), 3))` returns `[(0, 1, 2), (3, 4, 5), (6, 7, 8)]`
+
+    :param iterator: an iterator to divide in chunks
+    :param size: the size of each chunk
+    """
+    iterator = iter(iterator)
+    return iter(lambda: tuple(itertools.islice(iterator, size)), ())
+
+
+def safe_flush_to_disk(fhandle, real_path, use_fullsync=False):
+    """Tries to to its best to safely commit to disk.
+
+    Note that calling this is needed to reduce the risk of data loss due to, e.g., a power failure.
+    However, this call is typically expensive so should be called only if the guarantees are really needed, and
+    as few times as possible.
+
+    :param fhandle: an open file handle. Must support the `.fileno()` method and the `.flush()` method.
+    :param real_path: the real path of the file. It must be ideally the absolute path. It is used (on POSIX) to find
+        the parent folder and flush it too.
+    :param use_fullsync: on Mac OS, runs a FULLSYNC to really push data to disk.
+    """
+    fileno = fhandle.fileno()
+
+    # Flush buffers
+    fhandle.flush()
+
+    # Default fsync function, replaced on Mac OS X
+    _fsync_function = lambda fileno: os.fsync(fileno)  # pylint: disable=unnecessary-lambda
+
+    # Flush to disk
+    if hasattr(fcntl, 'F_FULLFSYNC') and (_MACOS_ALWAYS_USE_FULLSYNC or use_fullsync):
+        # This exists only on Mac OS X; See e.g. (link split on two lines, put them together):
+        # https://developer.apple.com/library/archive/documentation/System/Conceptual/ManPages_iPhoneOS/
+        #          man2/fsync.2.html
+        # that says:
+        # > For applications that require tighter guarantees about the integrity of
+        # > their data, Mac OS X provides the F_FULLFSYNC fcntl.  The F_FULLFSYNC
+        # > fcntl asks the drive to flush all buffered data to permanent storage.
+        # > Applications, such as databases, that require a strict ordering of writes
+        # > should use F_FULLFSYNC to ensure that their data is written in the order
+        # > they expect.  Please see fcntl(2) for more detail.
+        # Replace the _fsync_function
+        _fsync_function = lambda fileno: fcntl.fcntl(fileno, fcntl.F_FULLFSYNC)  # pylint: disable=no-member,useless-suppression
+    else:
+        # In general this is the function to call
+        _fsync_function(fileno)
+
+    # Flush also the parent directory on posix, see e.g.
+    # https://blog.gocept.com/2013/07/15/reliable-file-updates-with-python/
+    if os.name == 'posix':
+        dirfd = os.open(os.path.dirname(real_path), os.O_DIRECTORY)
+        # Also here call the correct fsync function
+        _fsync_function(dirfd)
+        os.close(dirfd)
+
+
+def compute_hash_and_size(stream, hash_type):
+    """Given a stream and a hash type, return the hash key (hexdigest) and the total size.
+
+    :param stream: an open stream
+    :param hash_type: the string with a name of a valid hash type
+    :return: a tuple with ``(hash, size)`` where ``hash`` is the hexdigest and ``size`` is the size in bytes
+    """
+    _hash_chunksize = 524288
+    hasher = get_hash(hash_type)()
+
+    # Read and hash all content
+    size = 0
+    while True:
+        next_chunk = stream.read(_hash_chunksize)
+        if not next_chunk:
+            # Empty returned value: EOF
+            break
+        hasher.update(next_chunk)
+        size += len(next_chunk)
+
+    return hasher.hexdigest(), size
